--------------------------------------------------------------------------------
-- (c) by c't magazin 
--------------------------------------------------------------------------------
--
-- Design name : CT_Lab_FPGA
--
-- Filename : Timing_Gen.vhd
-- Version : 1.0
-- Date: 25.7.2008
--
-- Engineer : Klaus Philipp
--
--------------------------------------------------------------------------------
-- Timing generator
--
-- The timing generator generates a one second pulse and the control signals
-- for two 32 bit counters and a multiplexer
--
--------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
--
-- The Unisim Library is used to define Xilinx primitives. It is also used during
-- simulation. The source can be viewed at %XILINX%\vhdl\src\unisims\unisim_VCOMP.vhd
--
library unisim;
use unisim.vcomponents.all;
--

entity Timing_Gen is
   port ( CLK        : in    std_logic; 
          FREQ_IN    : in    std_logic; 
          AB_SWITCH  : inout   std_logic; 
          CE_A       : inout   std_logic; 
          CE_B       : inout   std_logic; 
          CLR_A      : out   std_logic; 
          CLR_B      : out   std_logic; 
          FREQ_OUT_A : out   std_logic; 
          FREQ_OUT_B : out   std_logic;
          INTR_ACK   : in    std_logic;
          RESET      : in    std_logic;			 
          INTR       : out   std_logic);
end Timing_Gen;

----------------------------------------------------------------------------------------------------------------------------------
-- Second Timer and Circuit Control
--
-- This circuit generates a 1 second timer which is used to  
-- switch the enable from one counter to the other. 
--
----------------------------------------------------------------------------------------------------------------------------------

architecture BEHAVIORAL of Timing_Gen is

signal second_count       : integer range 0 to 49999999 :=0;
signal second_pulse       : std_logic;

signal ab_switch_delay    : std_logic_vector(3 downto 0);
signal counter_clear_delay: std_logic_vector(2 downto 0);

signal intr_internal      : std_logic;

begin

  second_logic: process(CLK, INTR_ACK, RESET, second_count)
  begin
    if rising_edge (CLK)
	 then
	   if RESET = '1'
		then
		  ab_switch <= '0';
		  second_count <= 0;
		else
        if second_count=49999999       --divide by 50,000,000 is 1s
--        if second_count=505            --divide by 504 for test purposes
		  then      
          second_count <= 0;
          second_pulse <= '1';
        else
          second_count <= second_count + 1;
          second_pulse <= '0';
        end if;
		
      -- counter selection switch toggles each second
		
        if second_pulse='1'
		  then
          ab_switch <= not ab_switch;
        end if;
		end if;
    end if;
  end process second_logic;

-- generate processor interrupt and wait for an acknowledgement

   gen_intr : process (CLK, RESET, second_pulse, INTR_ACK)
	
	begin
	
	 if rising_edge (CLK)
	 then
		if RESET = '1'
		then
		  intr_internal <= '0';
		else
		  if second_pulse='1'
		  then
			 intr_internal <= '1';
		  elsif INTR_ACK='1'
		  then
			 intr_internal <= '0';
		  else
		    intr_internal <= intr_internal;
		  end if;
		end if;
		INTR <= intr_internal;
	 end if;
   end process gen_intr;
  
  
  counter_control: process(FREQ_IN)
  begin
    if FREQ_IN'event and FREQ_IN='1' then

      -- Four stage register to ensure synhchronisation of control between clock domains

      ab_switch_delay <= ab_switch_delay(2 downto 0) & ab_switch;
      
      -- Separate enables form switch between counters when control is consistantly high or low.

      case ab_switch_delay(3 downto 1) is

        when "000" => CE_A <= '1';
                      CE_B <= '0';
							 
		  when "001" => CE_A <= CE_A;
                      CE_B <= CE_B;

		  when "110" => CE_A <= CE_A;
                      CE_B <= CE_B;

        when "111" => CE_A <= '0';
                      CE_B <= '1';

        when others => CE_A <= CE_A;
                       CE_B <= CE_B;

      end case;

    end if;
  end process counter_control;
  
  counter_clear_control: process(CLK, AB_SWITCH)
  begin
    if rising_edge (CLK)
	 then

      -- Two stage register to generate the signals to clear the counters

      counter_clear_delay <= counter_clear_delay(1 downto 0) & ab_switch;

      case counter_clear_delay(2 downto 1) is
							 
		  when "01" => CLR_A <= '0';
							CLR_B <= '1';

		  when "10" => CLR_A <= '1';
							CLR_B <= '0';

        when others => CLR_A <= '0';
                       CLR_B <= '0';

      end case;

    end if;
  end process counter_clear_control;

  -- 
  -- Global buffer on selected source 
  --

  buffer_clkin1: BUFG
  port map( O => FREQ_OUT_A,
            I => FREQ_IN);
	
  buffer_clkin2: BUFG
  port map( O => FREQ_OUT_B,
            I => FREQ_IN);
			 
end BEHAVIORAL;
