--------------------------------------------------------------------------------
--	DIGITALER SINUS-GENERATOR
--
-- Modulname:			sinusgenerator.vhd
-- Projektname:		funktionsgenerator.ise
-- Ersteller:			Thomas Brunnengrber, erweitert von Carsten Meyer
--	Firma:				Privat
-- Erstellt:			05.03.2006
-- Zielsystem:			Spartan 3
-- Tool versions:		XILINX ISE 8.1
-- Revision:			1.0
-- Abhngigkeiten:	Keine
-- 
--------------------------------------------------------------------------------
-- Beschreibung:
--
-- Dieses Modul ist ein generisches Basismodul zur Erstellung von digitalen
-- Funktionsgeneratoren (hier speziell Sinus).
-- 
-- Das erste Viertel einer Sinuswelle wird als Tabelle in einem RAM-Block des
-- Spartan3 abgelegt. Bei einer Breite eines Tabelleneintrags w_tabeintrag von
-- 16 bit ergibt sich damit eine Tabellenlnge l_tab von 1024 Eintrgen pro
--	BRAM.
-- Die Phasendifferenz zwischen zwei Eintrgen betrgt somit 360/4096.
-- Die tatschliche Auflsung betrgt damit etwa 11,3 bit bei einer Sinus-
-- Funktion. Wenn dies nicht ausreicht, mssen die Funktionswerte auf mehrere 
-- BRAMs verteilt werden.
--
-- In den Bereichen 90-180 und 270-360 werden die Adressen invertiert.
-- Im Bereich 180-360 mssen die Funktionswerte "negiert" werden. Wie diese
-- "Negation" aussieht, ist abhngig vom tatschlich verwendeten DAC. Fordert der
-- DAC Zweier-Komplemente, so kann statt dessen eine schnellere und logiksparende
-- Invertierung aller Bits angewandt werden, da die letzten 4 Bit der
-- Tabelleneintrge aufgrund der durch die Tabellenlnge effektiv auf 11,3 bit
-- begrenzten Auflsung letztlich nicht zur Verfgung stehen und daher der durch
-- die simple Invertierung begangene Fehler nicht von Bedeutung ist.
--
-- Bestimmung von Takt und Lnge des Phasenzhlers:
--
--	Taktfrequenz: f_takt
--	Phasenzhler: phase
--	Bitbreite des Phasenzhlers : w_phase bit
--	Lnge des Phasenzhlers: l_phase = 2^w_phase
--	Bei jedem Takt zum Phasenzhler addierter Wert: d_phase
--	=> Kleinste einzustellende Frequenz: f_min = f_takt / l_phase
--	=> Generierte Frequenz: f_gen = f_min * d_phase
--	=> oder d_phase = f_gen / f_min
--	wobei f_takt und l_phase und damit f_min beim Design festzulegende Konstanten sind.
--	Gibt man f_min und f_takt vor, so ergibt sich die Lnge des Phasenzhlers zu
--	=> l_phase = f_min / f_takt
--	Der Maximalwert d_phase_max von d_phase ergibt sich aus der maximal zu generierenden 
--	Frequenz f_gen_max zu
--	=> d_phase_max = f_gen_max / f_min
--
--	Ganz einfaches Beispiel:
--	w_phase = 8 bit
--	=> l_phase = 2^8 = 256 
--	=> phase = 0..255
--	f_takt = 128 Hz
--	=> f_min = 128 Hz / 256 = 0.5 Hz
--	f_gen = 1 Hz soll eingestellt werden
--	=> d_phase = 1 Hz / 0.5 Hz = 2
--
--	Konkretes Beispiel:
--	Forderungen: 
--		f_min = 0.01 Hz
--		f_max = 10 MHz
--		etwa 2-fach Oversampling bei 10 MHz, damit analoge Filterung einfacher wird
-- 	=> f_takt ca. 40 MHz
-- 
--	=> l_phase ca. 4e7 / 0.01 = 4e9 => w_phase = 32 bit => l_phase = 2^32 = 4294967296
--	=> f_takt = l_phase * f_min = 42949672.96 Hz = 4194304 Hz * 32/10 * 32/10 
--	=> Quarz mit 2^22 Hz = 4194304 Hz und zwei nachgeschaltete DCM mit jeweils 
--	CLKFX_MULTIPLY = 32 und CLKFX_DIVIDE = 10
--	Achtung: Wird der DCM mit solch niedriger Eingangs-Frequenz betrieben, drfen die
-- PLL-Ausgnge nicht genutzt werden!
--	Alternativ: 50 MHz Quarz * 17 * 24 / 19 / 25 = 42947368 Hz
--	=> d_phase_max = 1e7 / 0.01 = 1e9 => w_d_phase = 30 bit
--	=> f_gen_max = 2^30 * 0.01 = 10737418.23 Hz (einstellbar)
-- Ausgabe erfolgt mit x8000 Offset fr DAC
-- Pegeleinstellung ebenfalls mit x8000 Offset, x8000= Null, xFFFF = Max., x0000 = Max. invertiert
--------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

library UNISIM;
use UNISIM.VComponents.all;

entity triangle_generator is
	generic (
		w_phase : integer := 32;
		w_d_phase : integer := 32;
		w_dac_wert : integer := 16;
		w_dac_addr : integer := 16 );
	port (
		takt : in std_logic;
		d_phase : in std_logic_vector ( w_d_phase-1 downto 0 );
		dac_pegel : in std_logic_vector ( 15 downto 0 );
--		neustart : in std_logic;	-- Trigger bei	steigender Flanke
--		start_phase : in std_logic_vector ( w_phase-1 downto 0 ); 
-- bernahme synchron mit takt, wenn neustart = '1'
		dac_wert : out std_logic_vector ( w_dac_wert-1 downto 0 ) -- Tabelleneintrag
);
end entity triangle_generator;

architecture arch1 of triangle_generator is


	signal MultA : std_logic_vector ( 17 downto 0 ); 
	signal MultB : std_logic_vector ( 17 downto 0 ); 
	signal MultP : std_logic_vector ( 35 downto 0 ); 
	signal halbwelle : std_logic;

begin
   -- MULT18X18: 18 x 18 signed asynchronous multiplier
   --            Virtex-II/II-Pro, Spartan-3
   -- Xilinx HDL Language Template, version 10.1.3
   
   MULT18X18_inst : MULT18X18
   port map (
      P => MultP,    -- 36-bit multiplier output
      A => MultA,    -- 18-bit multiplier input
      B => MultB     -- 18-bit multiplier input
   );
	

	set_phase : process ( takt )
		variable phase : unsigned ( w_phase-1 downto 0 );
			-- Lnge: nur Viertelwelle in RAM gespeichert => index = 0..4*iadr-1 
			-- => Index-Breite w_itab = w_adr+2

	begin
		if rising_edge( takt ) then
			halbwelle <= std_logic( phase( w_phase-1 ) );
			phase := phase + unsigned( d_phase );
		end if;
		if halbwelle = '1' then
			MultA(17 downto 2) <= std_logic_vector(phase(30 downto 15)) xor x"7FFF";
		else
			MultA(17 downto 2) <= std_logic_vector(phase(30 downto 15)) xor x"8000";
		end if;
		MultB(17 downto 2) <= dac_pegel;
	end process;
-- Absolutwerte fr DAC
	dac_wert <= MultP(34 downto 19);						-- 2. Halbwelle mit Komplement-Korrektur

end arch1;
